<style>
  body { 
    font-family: Arial, sans-serif;
    font-size: 11px; 
  }
  
  table, thead, tbody, tr, td, th { 
    font-size: inherit; 
    font-family: inherit; 
  }

  #main { display:table-cell; }
  
  #index { border-bottom: solid 1px #777777; }

  #index table td { padding-left: 5px; }

  #graph { float: left; }

  #list { float: left; }

  #list table { margin-left: auto; margin-right: auto; }

  #list table td { padding-left: 5px; }

  #list table td.name, th.name { text-align: left; }

  #list table td.size, th.size { text-align: right; }

  #tooltip { 
     display: none; 
     position: absolute; 
     background-color: white; 
     border: solid 1px black; 
     padding: 3px; 
  }
</style>


